/* ftest.c */
/* Test floating-point, fixed-point, and microEuler modes. */
/* Uses one header and one library (vstrack1). */
/* Simple test that prints data to screen. */
/* Created  5/14/95 */
/* Ken Hawkins */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include "vstrack1.h"

long ac;
char ** av;

long checkParm(char * parm) {
long i;
  for(i = 1; i < ac; i++) if(!strcmpi(av[i],parm)) return i;
  return 0; // parm not found
} // checkParm

void main(int argc,char ** argv) {
TrackerData td;
TrackerStatus ts;
TrackerPort trackerPort;
float cy,ey,dy;
float gx,gy,mx,my,mz,sinp,sinr,cosp,cosr,rmx,rmy,rmz,fyaw,fcdy;
long i,baud,remap,cmode,emode;

// Setup global arg vals for checkParm().
  ac = argc;
  av = argv;

  if (checkParm("-?") || checkParm("/?") || checkParm("?") ||
    checkParm("-h") || checkParm("/h") || checkParm("h")) {
    printf("\n\n\nUsage:\n");
    printf("  ftest [-baud #] [-com #] [-remap] [-c | -e]\n");
    printf("    -baud #  : Baud rate: 1200, 2400, 4800, 9600, or 19200.\n");
    printf("    -com #   : Com port 1-4.\n");
    printf("    -remap   : Remap yaw quadrant for older tracker firmware.\n");
    printf("    -c or -e : -c for cooked mode only, -e for euler mode only.\n");
    printf("Example:\n  ftest -baud 19200 -com 2\n");
    printf("The default values are: com = search, baud = 9600.\n");
    printf("    while running, q = quit.\n");
    printf("\n\nDisplay Legend:\n\n");
    printf("FY = floating-point cooked yaw.\n");
    printf("CY = fixed-point cooked yaw.\n");
    printf("EY = euler yaw computed on H68C11.\n");
    printf("DY = delta yaw between fixed-point cooked and micro euler: abs(CY-EY).\n");
    printf("FCDY = delta yaw between floating- and fixed-point cooked: abs(FY-CY).\n");
    exit(0);
  } // if

  cmode = checkParm("-c") != 0;
  emode = checkParm("-e") != 0;
  if (cmode && emode) {
    printf("\n-c and -e are mutually exclusive, pick only one.\n");
    exit(0);
  } // if

  remap = checkParm("-remap") != 0;
  if (remap) printf("\nYaw quadrant will be remapped.\n");

  if (i=checkParm("-baud")) {
    if (!sscanf(av[i+1],"%ld",&baud)) {
      printf("A valid baud rate must be specified after the -baud argument.\n");
      printf("For example: -baud 19200\n");
      exit(1);
    } // if
    if (baud != 1200 && baud != 2400 && baud != 4800
     && baud != 9600 && baud != 19200) {
      printf("Baud must be 1200, 2400, 4800, 9600, or 19200.\n");
      exit(1);
    } // if
  } else {
    baud = 9600;
  } // if

  if (!(i = checkParm("-com"))) {
    TrackerPort trackerPort;
    printf("\nSearching for a Tracker: "); fflush(stdout);
    ts = initTrackerSearch(&td,&trackerPort,baud,timerSecs(2));
    printf("%s.\n",trackerInfo(ts));
    if (ts != TS_OK) {
      closeTracker(&td);
      exit(1);
    } // if
  } else {
    TrackerPort tport;
    long port;

    if (!sscanf(av[i+1],"%ld",&port)) {
      printf("A COM port (1-4) must be specified. Example: -com 1\n");
	  exit(1);
	} // if

    switch (port) {
      case 1: tport = TP_COM1; break;
      case 2: tport = TP_COM2; break;
      case 3: tport = TP_COM3; break;
      case 4: tport = TP_COM4; break;
      default:
        printf("Invalid COM port: %ld\n",port);
        exit(1);
    } // switch

    printf("Initializing Tracker on COM %ld.\n",port);

    ts = initTracker(&td,tport,baud,timerSecs(2));
    printf("%s.\n",trackerInfo(ts));
    if (ts != TS_OK) {
      closeTracker(&td);
      exit(1);
    } // if

  } // if

  while (1) {
    float y,p,r;
    static long inited=0;

    if (kbhit() && getch() == 'q') break;

// Cooked data
    if (cmode || !emode) {

      if (!inited || !(cmode || emode)) {
        inited = 1;
        ts = sendTrackerCMD(&td,"!M1,P,B\r",timerSecs(2));
        if (ts != TS_OK) {
          printf("%s.\n",trackerInfo(ts));
        } // if
      } // if

      requestTrackerData(&td);

      ts = readTracker(&td,timerSecs(1));
      if (ts != TS_OK) {
        printf("\n%s.\n",trackerInfo(ts));
        ts = resetTracker(&td,timerTSecs(1));
        if (ts != TS_OK) {
          printf("\n%s.\n",trackerInfo(ts));
        } // if
      } // if

      y = TOFLOAT(td.euler.y);
      p = TOFLOAT(td.euler.x);
      r = TOFLOAT(td.euler.z);

//    printf("C y %8.2f p %8.2f r %8.2f\n",y,p,r);

      cy = y;

// Compute yaw using floating point:
      gx = (float)td.gravimetric.x/16384.*M_PI;
      gy = (float)td.gravimetric.y/16384.*M_PI;

      mx = td.magnetic.x;
      my = td.magnetic.y;
      mz = td.magnetic.z;

      sinp = sin(gx);
      cosp = cos(gx);
      sinr = sin(gy);
      cosr = cos(gy);

      rmy = cosp*my - sinp*mz;
      rmz = sinp*my + cosp*mz;

      rmx = cosr*mx - sinr*rmy;

      fyaw = atan2(rmz,rmx)*180./M_PI;

// Compute difference between floating point and integer cooked modes:
      fcdy = abs(fyaw - cy);

    } // if

// Euler data
    if (!cmode || emode) {

      if (!inited || !(cmode || emode)) {
        inited = 1;
        ts = sendTrackerCMD(&td,"!M2,P,B\r",timerSecs(2));
        if (ts != TS_OK) {
          printf("%s.\n",trackerInfo(ts));
        } // if
      } // if

      requestTrackerData(&td);

      ts = readTrackerEuler(&td,timerSecs(1));
      if (ts != TS_OK) {
        printf("\n%s.\n",trackerInfo(ts));
        ts = resetTracker(&td,timerTSecs(1));
        if (ts != TS_OK) {
          printf("\n%s.\n",trackerInfo(ts));
        } // if
      } // if

      y = TOFLOAT(td.euler.y);
      p = TOFLOAT(td.euler.x);
      r = TOFLOAT(td.euler.z);

// Remap to same basis as Cooked mode:
      if (remap) {
        ey = -y + 270.;
        if (ey > 180.) ey -= 360.;
      } else {
        ey = y;
      } // if

      dy = abs(cy - ey);

    } // if

    if (cmode) {
      printf("FY %7.2f CY %7.2f FCDY %2.3f\n",fyaw,cy,fcdy);
    } else if (emode) {
      printf("EY %7.2f\n",ey);
    } else {
      printf("FY %7.2f CY %7.2f EY %7.2f DY %7.2f FCDY %2.3f\n",fyaw,cy,ey,dy,fcdy);
    } // if
  } // while

  closeTracker(&td);

} // main

/* test.c */

