/* serial.c
   Created 1/7/94
   Rewritten 6/8/94
   Copyright 1995 Ken Hawkins
   All Rights Reserved

   Modified by Paul J. Radek for BIMODAL support.
   Modified 2/11/95 by jcs for Single mode compile (Real or Protected mode only)
   Modified 2/24/95 by jcs for full DPMI and VMM compatibility.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#if !(defined(__WATCOMC__) && defined(__386__))
#include <conio.h>
#endif
#include <dos.h>
#include "serial.h"


/*******************************************************************************************************/
/* BIMODAL INTERRUPT HANDLER STUFF                                       */
/*******************************************************************************************************/

#ifdef __386__
void set_es(void);
#pragma aux set_es = "push ds", "pop es"
#endif

#define D32RealSeg(p)   ((((unsigned long)(p))>>4) & 0xFFFF)
#define D32RealOfs(p)   (((unsigned long)(p)) & 0x000F)

#define FLG_CHECK_CTS           (1<<0)
#define FLG_BLOCKED             (1<<1)
#define FLG_TX_PRIME            (1<<2)
#define FLG_USE_FIFO            (1<<3)

// #define FLG_CALLED (1<<4) is in serial.h

extern void __interrupt __far sbhInt( void );
extern void sbhEnd( void );

#if defined(__WATCOMC__) && defined(__386__)
// Make sure Watcom compiler creates correct symbols for register and stack
// based calling conventions for these assembly module labels.
#pragma aux sbhInt "*_"
#pragma aux sbhEnd "*_"
#endif

#if defined(__WATCOMC__) && defined(__386__)
// Use 32-bit protected mode bi-modal serial handler (both protected and real-mode handlers will be active).
#define BI_MODAL
// Don't #define USE_DPMI unless you are using DOS4G/W 1.97 or later:
// The DOS4G/W docs state that this method has less overhead. It generates
// a stack fault with version 1.96 under a DPMI host (e.g. Windows 3.1).
//#define USE_DPMI
#endif

/*******************************************************************************************************/
/*************************************** PC UART and Serial Code ***************************************/
/*******************************************************************************************************/

static unsigned short * serPort = (unsigned short *)0x400; /* Low memory port addresses */


#define PIC_CMD  0x20       /* Processor Interrupt Control register */
#define PIC_MASK 0x21       /* Interrupt enable mask register       */
#define EOI 0x20            /* End Of Interrupt command             */

/* Interrupt bit masks in the PC's 8259 Interrupt Controller        */
#define COM2_4_INT_BIT (1<<3) /* Com2 and com4 interrupt bit mask   */
#define COM1_3_INT_BIT (1<<4) /* Com1 and com3 interrupt bit mask   */

/* Interrupt vector offsets */
#define COM2_4_IRQ3_VEC 0x0B  /* Com2 and com4 */
#define COM1_3_IRQ4_VEC 0x0C  /* Com1 and com3 */

/* COM IO port register offsets */
#define TX_BUFFER     0  /* If bit 7 of line control register is 0 */
#define RX_BUFFER     0  /* If bit 7 of line control register is 0 */
#define DIVISOR_LO    0  /* If bit 7 of line control register is 1 */
#define DIVISOR_HI    1  /* If bit 7 of line control register is 1 */
#define INT_ENABLE    1  /* If bit 7 of line control register is 0 */
#define INT_ID        2
#define FIFO_16550    2  /* Only valid on 16550 UART */
#define LINE_CONTROL  3
#define MODEM_CONTROL 4
#define LINE_STATUS   5
#define MODEM_STATUS  6

/* FIFO Control Register bits (FCR register) */
#define FIFO_ENABLE (1<<0)
#define FIFO_RESET  (3<<1)
#define FIFO_RX1    (0<<6)
#define FIFO_RX4    (1<<6)
#define FIFO_RX8    (2<<6)
#define FIFO_RX14   (3<<6)
#define FIFO_INIT   (FIFO_RX14|FIFO_ENABLE)
#define FIFO_SET    (FIFO_RX4|FIFO_ENABLE) // |FIFO_RESET
#define FIFO_CLEAR  0

#define FIFO_WRITE_LEN 16

/* Interrupt enable register bit masks */
#define INT_EN_DELTA_MODEM (1<<3)
#define INT_EN_DELTA_LINE  (1<<2)
#define INT_EN_TX_EMPTY    (1<<1)
#define INT_EN_RX_READY    (1<<0)

#define INT_EN_SET         (INT_EN_TX_EMPTY|INT_EN_RX_READY|INT_EN_DELTA_MODEM)

/* Interrupt ID masks */
#define INT_ID_MODEM    (0<<1) /* Modem status register changed */
#define INT_ID_TX_EMPTY (1<<1) /* Transmitter holding buffer empty */
#define INT_ID_RX_READY (2<<1) /* Received data available */
#define INT_ID_LINE     (3<<1) /* Line status register changed */
#define INT_ID_PENDING  (1<<0) /* If zero, use bits 1 and 2 to check interrupt type, else no interrupt pending */
#define INT_ID_MASK     ((1<<3)-1)

/* Line control register bit masks */
#define LC_SET_LATCHES (1<<7) /* When set, allows setting of DIVISOR LO and HI at reg. loc. 0 and 1 */
#define LC_BREAK_ON    (1<<6) /* When set, enables break control */
#define LC_OVRD_PARITY (1<<5) /* When set, forces parity bit to 1, else 0 */
#define LC_SEL_PARITY  (1<<4) /* When set, use even parity, else odd */
#define LC_EN_PARITY   (1<<3) /* When set, enables parity, else disable */
#define LC_STOP_BITS   (1<<2) /* When set, use 2 stop bits, else use 1 (if word len 5 and set, use 1.5 stop bits) */
#define LC_WORD_LEN_8  0x03   /* Word length: %00 = 5 bits, %01 = 6, %10 = 7, and %11 = 8 */

#define LC_INIT        (LC_WORD_LEN_8) /* No break, no parity, 1 stop bit, 8 data bits */

/* Line status register bit masks */
#define LS_TX_IDLE       (1<<6) /* Set if transmitter idle    */
#define LS_TX_HOLD_EMPTY (1<<5) /* Set if can accept new char */
#define LS_BREAK         (1<<4) /* Set if break received      */
#define LS_FRAMING_ERROR (1<<3) /* Set if framing error       */
#define LS_PARITY_ERROR  (1<<2) /* Set if parity error        */
#define LS_OVERRUN_ERROR (1<<1) /* Set if overrun error       */
#define LS_RX_DATA_READY (1<<0) /* Set if received data ready */

/* Modem status register bit masks */
#define MS_RLSD       (1<<7) /* Receiver line signal detect (CD) */
#define MS_RI         (1<<6) /* Ring Indicator                   */
#define MS_DSR        (1<<5) /* Data set ready                   */
#define MS_CTS        (1<<4) /* Clear to send                    */
#define MS_DELTA_RLSD (1<<3)
#define MS_DELTA_RI   (1<<2)
#define MS_DELTA_DSR  (1<<1)
#define MS_DELTA_CTS  (1<<0)

/* Modem control register bit masks */
#define MC_LB_TEST    (1<<4) /* When set enables loop back test mode */
#define MC_OUT2_SIG   (1<<3) /* When set enables PC COM interrupts   */
#define MC_OUT1_SIG   (1<<2) /* When set resets some internal modems */
#define MC_RTS        (1<<1) /* When set enables request to send     */
#define MC_DTR        (1<<0) /* When set enables data terminal ready */

/* ComRate is divided into max bps to get divisor for latches */
#define MIN_BPS 300
#define MAX_BPS 115200

#ifdef BI_MODAL

// The locking code is for VMM/DPMI compatibility.

long lockMemory(void * mem,unsigned long size) {
union REGS r;

  r.x.eax = 0x0600; /* DPMI lock linear region */
  r.x.ebx = ((unsigned long)mem) >> 16;
  r.x.ecx = ((unsigned long)mem) & 0xFFFF;
  r.x.esi = size>>16;
  r.x.edi = size&0xffff;
  int386 (0x31,&r,&r);

//  printf("Lock size %6d result %d\n",size,r.x.cflag == 0);

  return (r.x.cflag == 0);

} // lockMemory

long unlockMemory(void * mem,unsigned long size) {
union REGS r;

  r.x.eax = 0x0601; /* DPMI unlock linear region */
  r.x.ebx = ((unsigned long)mem) >> 16;
  r.x.ecx = ((unsigned long)mem) & 0xFFFF;
  r.x.esi = size>>16;
  r.x.edi = size&0xffff;
  int386 (0x31,&r,&r);

//  printf("Unlock size %6d result %d\n",size,r.x.cflag == 0);

  return (r.x.cflag == 0);

} // unlockMemory

#endif

void START_LOCK_CODE(void) {}

long input(long port)
{
	return (long)inp(port);
} /* input */

void output(long port,long val)
{
	outp(port,val);
} /* output */

void
txHandler(
	serialData      *sd
	)
{
    register unsigned char   ch, status, i;
	BIH_STRUCT      *bih;
	QUE             *q;
	int             cnt;

	if ( ! sd->port )
		return;

	if ( sd->checkCTS )
	{
		bih = sd->bih;
		status = input( sd->port + MODEM_STATUS );
		if ( status & MS_CTS )
		{
			bih->flags &= ~FLG_BLOCKED;
		}
		else
		{
			bih->flags |= FLG_BLOCKED;
			return;
		}
	}
	status = input( sd->port + LINE_STATUS );
	if ( ! ( status & LS_TX_HOLD_EMPTY ) )
		return;

	cnt = ( sd->fifoEnabled ) ? FIFO_WRITE_LEN : 1;
	q = sd->tx.q;
	for ( i = 0; q->count > 0 && i < cnt; i++ )
	{
        ch = *( sd->tx.que + q->head++ );
        if ( q->head == q->size )
			q->head = 0;
		q->count--;
		output( sd->port + TX_BUFFER, ch );
	}
}

void
comIntHandler(
	serialData      *sd
	)
{
    register unsigned char   int_id,status,ch;
	QUE             *q;

#ifdef __386__
    set_es();
#endif
	if ( sd->port )
	{
		sd->bih->flags |= FLG_CALLED;
		for ( ;; )
		{
			int_id = input( sd->port + INT_ID ) & INT_ID_MASK;
			if ( int_id & INT_ID_PENDING )
				break; /* (no interrupts yet) */

			switch( int_id )
			{
				case INT_ID_RX_READY: /*** Read ***/
					for ( ;; )
					{
						status = input( sd->port + LINE_STATUS );
						if ( ! ( status & LS_RX_DATA_READY ) )
							break;

						ch = input( sd->port + RX_BUFFER ); /* Read char and clear int */
						q = sd->rx.q;

						if ( q->count < q->size )
						{
                            *( sd->rx.que + q->tail++ ) = ch;
                            if ( q->tail == q->size )
								q->tail = 0;
							q->count++;
						}
					}
					break;

				case INT_ID_TX_EMPTY: /*** Write ***/
					txHandler( sd );
					break;

				case INT_ID_MODEM:
					status = input( sd->port + MODEM_STATUS );
					if ( sd->checkCTS )
					{
						if ( status & MS_CTS )
						{ /* Clear To Send more characters */
							if ( sd->bih->flags & FLG_BLOCKED )
							{
								txHandler( sd );
							}
						}
						else
						{
							sd->bih->flags |= FLG_BLOCKED;
						}
					}
					break;

				case INT_ID_LINE:
					input( sd->port + LINE_STATUS );
					break;
			}
		}
	}
	output( PIC_CMD, EOI );
}

static serialData * sdcom[4]={0};

void __interrupt __far com_int_handler0(void)
{
	comIntHandler(sdcom[0]);
}
void __interrupt __far com_int_handler1(void)
{
	comIntHandler(sdcom[1]);
}
void __interrupt __far com_int_handler2(void)
{
	comIntHandler(sdcom[2]);
}
void __interrupt __far com_int_handler3(void)
{
	comIntHandler(sdcom[3]);
}

static s_ifunc int_handler[ 4 ] =
{
	com_int_handler0,
	com_int_handler1,
	com_int_handler2,
	com_int_handler3
};

void END_LOCK_CODE(void){}

long setBPS(serialData * sd,long bpsRate)
{
	long divisor,actual;
	if (!sd->port) return 0;

	if (bpsRate < MIN_BPS)
	{
		bpsRate = MIN_BPS;
	} else if (bpsRate > MAX_BPS)
	{
		bpsRate = MAX_BPS;
	} /* if */

	divisor = MAX_BPS/bpsRate;
	actual  = MAX_BPS/divisor;

	_disable();
	output(sd->port+LINE_CONTROL,LC_SET_LATCHES|LC_WORD_LEN_8);
	output(sd->port+DIVISOR_HI,divisor>>8);
	output(sd->port+DIVISOR_LO,divisor&0xff);
	output(sd->port+LINE_CONTROL,LC_WORD_LEN_8);
	_enable();

	return actual;

} /* setBPS */

void
destroySerQueue(
	serQueue        *sq
	)
{

#ifdef BI_MODAL

 union REGS      r;

	if ( sq->que )
	{
		r.x.eax = 0x101;                /* DOS free memory */
		r.w.dx = sq->selector;
		int386( 0x31, &r, &r );

// Unlock queue memory.
        unlockMemory(sq->que,sq->q->size);

    }

#else

  if (sq->que) {
    free(sq->que);
    sq->que = 0;
  }

#endif

  memset( sq, 0, sizeof( serQueue ) );
}

int
createSerQueue(
	QUE             *que,
	serQueue        *sq,
	int             size
	)
{
	union REGS      r;

	sq->q = que;
	if  ( sq->que == NULL )
	{
#ifdef BI_MODAL
		r.x.eax = 0x100;                         /* DOS allocate DOS memory */
		r.x.ebx = ( size + 15 ) >> 4;   /* Number of paragraphs requested */
		int386( 0x31, &r, &r );

		if ( r.x.cflag )  /* Failed */
			return 0;

		sq->que = ( char * )((r.x.eax & 0x0000FFFF ) << 4 );
		sq->selector = r.w.dx;
		que->seg = r.w.ax;

// Lock queue memory.
        lockMemory(sq->que,size);
#else
		sq->que = (char *)calloc(1,size);
		if (!sq->que) return 0;
#endif
	}
	que->head = 0;
	que->tail = 0;
	que->count = 0;
	que->size = size;

	return 1;
}



int
createBimodalHandler(
	serialData      *sd
	)
{
	unsigned long   size;
	union REGS      r;
	BIH_STRUCT      *bih;

#ifdef BI_MODAL

	size = ( unsigned long ) sbhEnd - (unsigned long) sbhInt;

    r.x.eax = 0x100;              /* DOS allocate DOS memory */
	r.x.ebx = ( size + 15 ) >> 4; /* Number of paragraphs requested */
	int386( 0x31, &r, &r );

	if ( r.x.cflag )  /* Failed */
		return 0;

	bih = sd->bih = ( BIH_STRUCT * )((r.x.eax & 0x0000FFFF ) << 4 );
	sd->bihSel = r.w.dx;

	// Copy real mode code down to DOS memory.
	memcpy( bih, ( void * ) sbhInt, size );

// Lock real mode code.
    lockMemory(bih,size);

#else

	bih = sd->bih = (BIH_STRUCT *)calloc(1,sizeof(*bih));
	if (!bih) return 0;

#endif

	bih->port = sd->port;
	bih->flags = 0;
	memset( &bih->rxq, 0, sizeof( QUE ) );
	memset( &bih->txq, 0, sizeof( QUE ) );

	return 1;
}

/********************************************************************/
/************************** Serial Functions ************************/
/********************************************************************/

long
initSerial(
	serialData      *sd,
	long            port,
	long            flags,
	long            txQsize,
	long            rxQsize
	)
{
	long            fcr,lcr,div_lo,div_hi;
	union REGS      r;

	deInitSerial( sd );

	if ( port == SERIAL_PORT_COM1 || port == SERIAL_PORT_COM3 )
	{
		sd->intvec = COM1_3_IRQ4_VEC;
		sd->intbit = COM1_3_INT_BIT;
		sd->port = serPort[port];
	}
	else if ( port == SERIAL_PORT_COM2 || port == SERIAL_PORT_COM4 )
	{
		sd->intvec = COM2_4_IRQ3_VEC;
		sd->intbit = COM2_4_INT_BIT;
		sd->port = serPort[port];
	}
	else
	{
		sd->port = 0;
	}
	if ( ! sd->port || sdcom[ port ] != 0 )
	{
		deInitSerial( sd );
		return SER_NO_PORT; /* Already in use */
	}
	sd->portNum = port;
	sdcom[ port ] = sd;

	if ( ! createBimodalHandler( sd ) ||
		 ! createSerQueue( &sd->bih->txq, &sd->tx, txQsize ) ||
		 ! createSerQueue( &sd->bih->rxq, &sd->rx, rxQsize )
	   )
	{
		deInitSerial( sd );
		return SER_NO_MEMORY;
	}

	_disable();

	lcr = inp( sd->port + LINE_CONTROL );
	sd->prevLineControl = lcr;
	lcr |= LC_SET_LATCHES;
	output( sd->port + LINE_CONTROL, lcr );
	div_hi = inp( sd->port + DIVISOR_HI );
	div_lo = inp( sd->port + DIVISOR_LO );

	if ( div_hi | div_lo )
		sd->prevBPS = MAX_BPS / ((div_hi << 8) | div_lo);
	else
		sd->prevBPS = 1200;

	lcr &= ~LC_SET_LATCHES;
	output( sd->port + LINE_CONTROL, lcr );

	sd->prevModemControl = inp( sd->port + MODEM_CONTROL );
	sd->prevPicMask      = inp( PIC_MASK );
	sd->prevIntEnable    = inp( sd->port + INT_ENABLE );
	sd->prevFifo         = inp( sd->port + FIFO_16550 );

	output( PIC_MASK, input( PIC_MASK ) | sd->intbit ); /* mask: Disable 8259 ints */
	output( sd->port + MODEM_CONTROL, 0 );              /* Disable com interrupts */
	output( sd->port + INT_ENABLE, 0 );                 /* Disable tx and rx ints */
	output( sd->port + FIFO_16550, FIFO_CLEAR );

// Chain serial handler(s) into system.

#ifndef BI_MODAL

    sd->prevint = _dos_getvect(sd->intvec);
    _dos_setvect( sd->intvec, int_handler[ port ] );

#else

    r.x.eax = 0x200;     /* DPMI get real mode vector */
    r.x.ebx = sd->intvec;
    int386(0x31,&r,&r);
    sd->wRmSel = (unsigned short) r.x.ecx;
    sd->wRmOfs = (unsigned short) r.x.edx;

#ifdef USE_DPMI

    r.x.eax = 0x0204;    /* DPMI get protected mode vector */
    r.x.ebx = sd->intvec;
    int386(0x31,&r,&r);
    sd->wPmSel = (unsigned short)r.x.ecx;
    sd->wPmOfs = r.x.edx;

    r.x.eax = 0x201;     /* DPMI set real mode vector */
    r.x.ebx = sd->intvec;
    r.x.ecx = D32RealSeg(sd->bih);
    r.x.edx = D32RealOfs(sd->bih);
    int386(0x31,&r,&r);

    r.x.eax = 0x205;     /* DPMI set protected mode vector */
    r.x.ebx = sd->intvec;
    r.x.ecx = (unsigned long)FP_SEG((void __far *)int_handler[port]);
    r.x.edx = FP_OFF((void __far *)int_handler[port]);
    int386(0x31,&r,&r);

#else
                         /* DOS setvect (Both protected and real) */
    sd->prevint = _dos_getvect(sd->intvec);
    _dos_setvect(sd->intvec,int_handler[port]);

    r.x.eax = 0x201;     /* DPMI set real mode vector */
    r.x.ebx = sd->intvec;
    r.x.ecx = D32RealSeg(sd->bih);
    r.x.edx = D32RealOfs(sd->bih);
    int386(0x31,&r,&r);

#endif

// Lock protected mode code.
    lockMemory(START_LOCK_CODE,
      (unsigned long)END_LOCK_CODE-(unsigned long)START_LOCK_CODE);

#endif

// FIFO

	if ( flags & USE_16550 )
	{
		output( sd->port + FIFO_16550, FIFO_INIT );
		fcr = input( sd->port + FIFO_16550 );
		if ((fcr & 0xf8) == 0xc0 )
		{
			sd->fifoEnabled = 1;
            sd->bih->flags |= FLG_USE_FIFO;
            output( sd->port + FIFO_16550, FIFO_SET );
		}
		else
		{
            output( sd->port + FIFO_16550, FIFO_CLEAR );
		}

    }
	output( sd->port + INT_ENABLE, INT_EN_SET );                   /* Enable tx and rx ints and delta modem ints    */
	output( sd->port + MODEM_CONTROL, MC_OUT2_SIG | MC_RTS | MC_DTR ); /* Enable com interrupts, enable RTS, enable DTR */
	output( PIC_MASK, input(PIC_MASK) & (~sd->intbit) );           /* unmask: Enable 8259 ints                      */

	input( sd->port + INT_ID );       /* Dummy read to enable */
	input( sd->port + RX_BUFFER );    /* Dummy read to enable */
	input( sd->port + LINE_STATUS );  /* Dummy read to enable */
	input( sd->port + MODEM_STATUS ); /* Dummy read to enable */

	_enable();

	setRTS( sd, 1 );
	setDTR( sd, 1 );

	return SER_OK;
}

void deInitSerial(serialData * sd) {
union REGS r;

  if (sd->port) {
    _disable();

#ifdef BI_MODAL

#ifdef USE_DPMI

    if (sd->wPmSel || sd->wPmOfs) {

      r.x.eax = 0x205;      /* DPMI set protected mode vector */
      r.x.ebx = (unsigned long)sd->intvec;
      r.x.ecx = (unsigned long)sd->wPmSel;
      r.x.edx = sd->wPmOfs;
      int386(0x31,&r,&r);

    } // if

#else
      if (sd->prevint) _dos_setvect(sd->intvec,sd->prevint);
#endif

// Unlock protected-mode code.
    unlockMemory(START_LOCK_CODE,
      (unsigned long)END_LOCK_CODE-(unsigned long)START_LOCK_CODE);

    if (sd->bih) {
      r.x.eax = 0x201;    /* DPMI set real mode vector */
      r.h.bl  = sd->intvec;
      r.x.ecx = sd->wRmSel;
      r.x.edx = sd->wRmOfs;
      int386(0x31,&r,&r);
    } // if

    if (sd->bihSel) {
      r.x.eax = 0x101;      /* Free BIH memory */
      r.w.dx  = sd->bihSel;
      int386(0x31,&r,&r);
// Unlock real-mode code.
      unlockMemory(sd->bih,(unsigned long)sbhEnd-(unsigned long)sbhInt);
    } // if

#else
    if (sd->prevint) {
      _dos_setvect(sd->intvec,sd->prevint);

      if (sd->bih) {
        free(sd->bih);
        sd->bih = 0;
      } // if
    } // if
#endif

    output( sd->port + MODEM_CONTROL, 0 ); /* Disable com interrupts */
    output( sd->port + INT_ENABLE, 0 );    /* Disable tx and rx ints */

    output(sd->port + MODEM_CONTROL,sd->prevModemControl);
    if (sd->prevPicMask & sd->intbit) {
      output(PIC_MASK,input(PIC_MASK) | sd->intbit); // Mask and disable
    } // if
    output( sd->port + INT_ENABLE, sd->prevIntEnable );
    if ((sd->prevFifo & 0xf8) != 0xc0 && sd->fifoEnabled) {
      output(sd->port + FIFO_16550,FIFO_CLEAR);
    } // if

    setBPS(sd,sd->prevBPS);
    output(sd->port+LINE_CONTROL,sd->prevLineControl);

    destroySerQueue(&sd->tx);
    destroySerQueue(&sd->rx);

    sdcom[sd->portNum] = 0;
    memset(sd,0,sizeof(serialData));

    input( sd->port + INT_ID );
    input( sd->port + RX_BUFFER );
    input( sd->port + LINE_STATUS );
    input( sd->port + MODEM_STATUS );

    _enable();

  } // if
} // deInitSerial

long
writeSer(
	serialData      *sd,
	char            ch
	)
{
	QUE             *q;

	if ( sd->port )
	{
        q = sd->tx.q;

		if ( q->count == q->size )
			return 0;

		_disable();
		*( sd->tx.que + q->tail++ ) = ch;
		if ( q->tail == q->size )
			q->tail = 0;
		q->count++;
        txHandler( sd );
        _enable();
    }
	return 1;
}

void writeSerNoBuff(serialData * sd,char ch) {
QUE * q;
  if (sd->port) {
    q = sd->tx.q;
// Wait for tx queue to empty.
    while (q->count);
// Wait until tx register is available.
    while (!(input( sd->port + LINE_STATUS ) & LS_TX_HOLD_EMPTY));
// Write character.
    output( sd->port + TX_BUFFER, ch );
  } // if
} // writeSerNoBuff

/* Call readSer() only after rxBuffEmpty returns false */
long
readSer(
	serialData      *sd
	)
{
	long            ch;
	QUE             *q;

	if ( ! sd->port )
		return -1;

	q = sd->rx.q;
	if ( q->count <= 0 )
		return -1;

	_disable();
	ch = *( sd->rx.que + q->head++ );
	if ( q->head == q->size )
		q->head = 0;
	q->count--;
	_enable();

	return ch;
}

long
rxBuffEmpty(
	serialData      *sd
	)
{
	if ( ! sd->port )
		return 1;

	return ( sd->rx.q->count ) ? 0 : 1;
}

long
txBuffEmpty(
	serialData      *sd
	)
{
	if ( ! sd->port )
		return 1;

	if ( ! sd->tx.q->count )
	{
		return ( ( input( sd->port + LINE_STATUS ) & LS_TX_IDLE ) != 0 );
	}
	return 0;
}

void
flushRead(
	serialData      *sd
	)
{
	if ( sd->port )
	{
		_disable();
		sd->rx.q->count = 0;
		sd->rx.q->head = 0;
		sd->rx.q->tail = 0;
		input( sd->port + RX_BUFFER );
		_enable();
	}
}

void flushWrite(serialData * sd)
{
	if ( sd->port )
	{
		_disable();
		sd->tx.q->count = 0;
		sd->tx.q->head = 0;
		sd->tx.q->tail = 0;
		sd->bih->flags &= ~FLG_BLOCKED;
		_enable();
		while ( ! txBuffEmpty( sd ) )
			;
	}
}

void
flushSer(
	serialData      *sd
	)
{
	if ( ! sd->port )
		return;

	_disable();
	sd->tx.q->count = 0;
	sd->tx.q->head = 0;
	sd->tx.q->tail = 0;
	sd->rx.q->count = 0;
	sd->rx.q->head = 0;
	sd->rx.q->tail = 0;
	sd->bih->flags &= ~FLG_BLOCKED;
	input( sd->port + LINE_STATUS );
	input( sd->port + MODEM_STATUS );
	input( sd->port + INT_ID );
	input( sd->port + RX_BUFFER );
	_enable();
}

#define SETSERBIT(b) output(sd->port+MODEM_CONTROL,input(sd->port+MODEM_CONTROL)|(b))
#define CLRSERBIT(b) output(sd->port+MODEM_CONTROL,input(sd->port+MODEM_CONTROL)&(~(b)))
#define GETSERBIT(b) (input(sd->port+MODEM_STATUS)&(b))

void
setCheckCTS(
	serialData      *sd,
	long            checkCTS
	)
{
	if ( checkCTS )
	{
		sd->checkCTS = 1;
		sd->bih->flags |= FLG_CHECK_CTS;
	}
	else
	{
		sd->checkCTS = 0;
		sd->bih->flags &= ~FLG_CHECK_CTS;
	}
}

void
setDTR(
	serialData      *sd,
	long            state
	)
{
	if ( sd->port )
	{
		if ( state )
			SETSERBIT( MC_DTR ); /* Active high */
		else
			CLRSERBIT( MC_DTR );
	}
}

void
setRTS(
	serialData      *sd,
	long            state
	)
{
	if ( sd->port )
	{
		if ( state )
			SETSERBIT( MC_RTS ); /* Active high */
		else
			CLRSERBIT( MC_RTS );
	}
}

long
getDSR(
	serialData      *sd
	)
{
	if ( sd->port )
		return GETSERBIT(MS_DSR) != 0; /* Active high */
	else
		return 0;
}

long
getDCD(
	serialData      *sd
	)
{
	if ( sd->port )
		return GETSERBIT(MS_RLSD) != 0; /* Active high */
	else
		return 0;
}

long
uartType(
	serialData      *sd
	)
{
	if ( sd->port )
	{
		if ( sd->fifoEnabled )
			return UART_16550;
		else
			return UART_8250;
	}
	return UART_NONE;
}

long
serOpen(
	serialData      *sd
	)
{
	return sd->port != 0;
}

