/* vtmode.c */
/* Set all possible VIO tracker modes */
/* Created 1/4/95 */
/* Ken Hawkins */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "imath1.h"
#include "vector1.h"
#include "convert1.h"
#include "serial1.h"
#include "tracker.h"
#include "tcal.h"
#include "spacerat.h"

void vtmode(char * msg,int argc,char ** argv) {
TrackerData td;
TrackerStatus ts;
char cmd[256];

long comport,comrate,datamode,sendmode,format,magfilt,grvfilt;
long xsensi,ysensi,xthresh,ythresh;
long bpsrate;

  printf("\n");

  if (argc < 6 || argc > 12) {
    puts(msg);
    printf("         0     1     2     3     4         5             6            7\n");
    printf("args = [1-4] [0-4] [0-3] [C|P] [A|B] {[0-7] [0-7]} {[0-9] [0-9]  [0-9] [0-9]}\n",argv[0]);
    printf("0: ComPort [1-4]\n");
    printf("1: ComRate\n");
    printf("   0 = 1200, 1 = 2400, 2 = 4800, 3 = 9600, 4 = 19200\n");
    printf("2: Data Mode\n");
    printf("   0 = Raw, 1 = Cooked, 2 = Euler, 3 = Mouse\n");
    printf("3: Send Mode\n");
    printf("   C = Continuous, P = Polled\n");
    printf("   0-1: Mouse modes (0 = Delta, 1 = Joystick)\n");
    printf("4: Format\n");
    printf("   A = ASCII, B = Binary\n");
    printf("Optional (Filters must be specified if mouse mode settings used):\n");
    printf("5: Filters {[Magnetic] [Gravimetric]}\n");
    printf("Mouse Mode Settings:\n");
    printf("6: Sensitivity {[x] [y] \n");
    printf("7: Threshold    [x] [y]}\n");
    exit(0);
  } // if

  comport  = atoi(argv[1]);
  comrate  = atoi(argv[2]);

  switch (comport) {
    case 1: comport = TP_COM1; break;
    case 2: comport = TP_COM2; break;
    case 3: comport = TP_COM3; break;
    case 4: comport = TP_COM4; break;
    default:
      printf("Invalid ComPort: %d\n",comport);
      exit(0);
  } // switch

  switch (comrate) {
    case 0: bpsrate =  1200; break;
    case 1: bpsrate =  2400; break;
    case 2: bpsrate =  4800; break;
    case 3: bpsrate =  9600; break;
    case 4: bpsrate = 19200; break;
    default:
      printf("Invalid ComRate: %d\n",comrate);
      exit(0);
  } // switch

  datamode = atoi(argv[3]);
  sendmode = argv[4][0];
  format   = argv[5][0];

  if (datamode < 0 || datamode > 3) {
    printf("Invalid data mode: %d\n",datamode);
    exit(0);
  } // if

  if (sendmode != 'C' && sendmode != 'P' &&
    sendmode != '0' && sendmode != '1') {
    printf("Invalid send mode: %c\n",sendmode);
    exit(0);
  } // if

  if (format != 'A' && format != 'B') {
    printf("Invalid format: %c\n",format);
    exit(0);
  } // if

  if (argc == 7) {
    printf("Invalid argument format. Both filters must be specified.\n");
    exit(0);
  } // if

  if (argc >= 8) {
    magfilt = atoi(argv[6]);
    grvfilt = atoi(argv[7]);
    if (magfilt < 0 || magfilt > 7) {
      printf("Invalid magnetic filter setting: %d\n",magfilt);
      exit(0);
    } // if
    if (grvfilt < 0 || grvfilt > 7) {
      printf("Invalid gravimetric filter setting: %d\n",grvfilt);
      exit(0);
    } // if

    if (argc == 12) {
      xsensi  = atoi(argv[8]);
      ysensi  = atoi(argv[9]);
      xthresh = atoi(argv[10]);
      ythresh = atoi(argv[11]);

      if (xsensi < 0 || xsensi > 9) {
        printf("Invalid mouse x sensitivity: %d\n",xsensi);
        exit(0);
      } // if

      if (ysensi < 0 || ysensi > 9) {
        printf("Invalid mouse y sensitivity: %d\n",ysensi);
        exit(0);
      } // if

      if (xthresh < 0 || xthresh > 9) {
        printf("Invalid mouse x threshold: %d\n",xthresh);
        exit(0);
      } // if

      if (ythresh < 0 || ythresh > 9) {
        printf("Invalid mouse y threshold: %d\n",ythresh);
        exit(0);
      } // if
    } else if (argc != 8) {
      printf("Invalid argument format. All mouse options must be specified.\n");
      exit(0);
    } // if

  } // if

  switch (argc) {
    case 6:
      sprintf(cmd,"!M%d,%c,%c\r",datamode,sendmode,format);
      break;
    case 8:
      sprintf(cmd,"!M%d,%c,%c,%d,%d\r",datamode,sendmode,format,
        magfilt,grvfilt);
      break;
    case 12:
      sprintf(cmd,"!M%d,%c,%c,%d,%d,%d,%d,%d,%d\r",datamode,sendmode,format,
        magfilt,grvfilt,xsensi,ysensi,xthresh,ythresh);
      break;
    default:
      printf("Invalid argument format.\n");
      exit(0);
   } // switch

  printf("Initializing tracker on COM%d at %d bps: ",comport+1,bpsrate);

  ts = initTracker(&td,comport,bpsrate,timerSecs(2));
  printf(" %s.\n",trackerInfo(ts));
  if (ts != TS_OK) {
    closeTracker(&td);
    exit(0);
  } // if

  ts = sendTrackerCMD(&td,cmd,timerSecs(3));
  printf(" %s.\n",trackerInfo(ts));
  if (ts != TS_OK) goto done;

  delayTime(timerTSecs(3));

  requestTrackerData(&td);

  if (datamode == 3) {
    printf("Running SpaceRat. Move the tracker around to test. Press 'q' to quit.\n");
    spacerat(&td);
  } else {
    long startT;
    startT = getTime();
    while (rxBuffEmpty(td.sd)) {
      if (elapsedTime(startT,timerSecs(3))) {
        printf("The tracker didn't respond.\n");
        closeTracker(&td);
        exit(0);
      } // if
    } // while
  } // if

  printf("The tracker is ready.\n");

done:

  closeTracker(&td);

} // vtmode

/* vtmode.c */

